#include <iostream>
#include <fstream>
#include <vector>
#include <time.h>
#include <assert.h>
#include <stdlib.h>
#include <string.h>

using namespace std;

struct indx {
    int i,j;
};

struct cell {
    indx place;
    char p;
    bool is_stem;
};

//global variables
bool lattice[750][750]; //lattice
vector<cell> cells; //vector containing all cells present in the system
int N=750; //lattice size

int cc=0,csc=1;
vector<pair<int, int> > trajectory;

//simualtion specific parmeter
char pmax=10; //divisions before proliferative capacity exhaustion
double pDiv=1./24.; //division probability
double alpha=0.01; //spontaneous death probability
double ps=0.1; //probability of symmetric division
double pmig=10./24.; //probability of migration

indx returnEmptyPlace (indx place) {//FUNCTION RETTURNING RANDOM FREE SPOT (IF THERE IS ANY)
    vector<indx> sites;
    indx aux;
    
    //gathering all empty sites
    for (int i=-1; i<2; i++)
        for (int j=-1; j<2; j++) {
            aux.i=place.i+i; aux.j=place.j+j;
            if ((i!=0 || j!=0) && aux.i>=0 && aux.i<N && aux.j>=0 && aux.j<N && !lattice[aux.i][aux.j])
                sites.push_back(aux);
        }
    
    //picking free spot at random
    if (!sites.empty())
        return sites.at(rand() % sites.size());
    
    return place;
}


void initialize() {//FUNCTON INITIALIZING CA
    lattice[N/2][N/2]=1;
    indx primaryPlace={N/2, N/2};
    cell primaryCell={primaryPlace, pmax, true};
    cells.push_back(primaryCell);
}

void save_structure(int number){
	
	
	char tmp[50],tmp2[50];
	sprintf(tmp, "_%i.csv", number);
	strcat(tmp2,"struct");
    strcat(tmp2,tmp);
    ofstream file_op(tmp2);
    
    
	if (file_op) {
        file_op<<"x coord, y coord, is_stem, p, nFreeSpots"<<endl;
        vector<cell>::iterator it;
        
        file_op<< N <<", " << N << ", 0, 0, 0";
        
        for (it=cells.begin(); it!=cells.end(); it++) {
            file_op<<endl;
            file_op<< it->place.i << ", " << it->place.j << ", "<< it->is_stem <<", "<<(int)it->p<<", "<< (int)lattice[it->place.i][it->place.j];
        }
        file_op.close();
	}
}


void simulate(int nSteps) {//MAIN SIMUALTION PROCEDURE
    
    vector<cell> cellsTmp;
    int randPick;
    double fate;
    cell currCell;
    indx newPlace;
    
    for (int i=0; i<nSteps; i++) {
        if ((i % 24) == 0 ) {
            //cout << "Day no.: " << i/24 << ", populationSize: " << cells.size() << endl;
            //cout << cells.size() << " ";
            pair<int, int> currSt (cc,csc);
            trajectory.push_back(currSt);
        }
        
        if ((i % (24*30)) == 0 ) {
            save_structure(i);
        }
        
       while (!cells.empty()) {
            randPick=(rand() % cells.size()); //choose which cell to pick
            currCell=cells.at(randPick); //pick the cell
            cells.erase(cells.begin()+randPick); //remove cell from current stack
           
            fate=((double)rand()/(double)RAND_MAX);
           
            if (fate<alpha && !currCell.is_stem) { //spontaneous death occurs
                lattice[currCell.place.i][currCell.place.j]=0;
                cc--;
            } else {
                newPlace = returnEmptyPlace(currCell.place);
                if (newPlace.i==currCell.place.i && newPlace.j==currCell.place.j) {//no available space around
                    cellsTmp.push_back(currCell);
                } else {
                    fate=((double)rand()/(double)RAND_MAX);
                    if (fate<pDiv) {//division
                        cell newCell=currCell;
                        newCell.place=newPlace;
                        if (currCell.is_stem) {//if stem cell is divding
                            fate=((double)rand()/(double)RAND_MAX);
                            if (fate>ps) {//asymmetric division
                                newCell.is_stem=false;
                                cc++;
                            } else {
                                csc++;
                            }
                            cellsTmp.push_back(currCell);
                            cellsTmp.push_back(newCell);
                            lattice[newPlace.i][newPlace.j]=true;
                        } else if (currCell.p==0) {//capacity exhausted
                            lattice[currCell.place.i][currCell.place.j]=false;
                            cc--;
                        } else {
                            newCell.p-=1; currCell.p-=1;
                            cellsTmp.push_back(currCell);
                            cellsTmp.push_back(newCell);
                            lattice[newPlace.i][newPlace.j]=true;
                            cc++;
                        }
        
                    } else if (fate<pmig) {//migration
                        lattice[currCell.place.i][currCell.place.j]=false;
                        lattice[newPlace.i][newPlace.j]=true;
                        currCell.place=newPlace;
                        cellsTmp.push_back(currCell);
                    } else { //do nothing
                        cellsTmp.push_back(currCell);
                    }
                }
             
            }
     
        }

        cells.swap(cellsTmp);
    }
}

int main() {
    clock_t start, stop;
    double t = 0.0;
    
    srand(time(NULL)); //initialize random number generator
    
    initialize(); //initialize CA
    
    assert((start = clock())!=-1);
    simulate(24*180);
    stop = clock();
    t = (double) (stop-start)/CLOCKS_PER_SEC;
    //cout << endl << "Run time: " << t << endl;
    ofstream outfile;
    
    outfile.open("outData.txt", std::ios_base::app);
    for (int i=0; i<trajectory.size(); i++)
        outfile << trajectory.at(i).first << " ";
    
    outfile << endl;
    
    for (int i=0; i<trajectory.size(); i++)
        outfile << trajectory.at(i).second << " ";
    
    outfile << endl;
    outfile.close();
    
    outfile.open("outRunTimes.txt", std::ios_base::app);
    
    
    outfile << t << endl;
    
    outfile.close();
    
    save_structure(1);
    cout << "done!" << endl;
    return 0;
}
